# -*- coding:utf-8 -*-
# Author:      zhousf
# Description:  加密工具
# pip install pycryptodome -i https://pypi.mirrors.ustc.edu.cn/simple/
import base64
import hashlib
import random
import string
from Crypto.Cipher import AES


class AESUtil(object):
    def __init__(self, key, mode):
        self.key = key
        self.mode = mode
        self.aes = AES.new(key=self.zero_padding_16(self.key), mode=self.mode)

    @staticmethod
    def zero_padding_16(text):
        """
        128位，0填充
        :param text:
        :return:
        """
        while len(text) % 16 != 0:
            text += '\0'
        return str.encode(text)

    def encrypt(self, text):
        """
        加密
        :param text:
        :return: 加密后的字符串
        """
        text_byte = self.zero_padding_16(text)
        return str(base64.encodebytes(self.aes.encrypt(text_byte)), encoding='utf8').replace('\n', '')

    def decrypt(self, text):
        """
        解密
        :param text:
        :return: 解密后的字符串
        """
        text_byte = bytes(text, encoding='utf8')
        return str(self.aes.decrypt(base64.decodebytes(text_byte)).rstrip(b'\0').decode("utf8"))


def md5(text):
    return hashlib.md5(text.encode("utf8")).hexdigest()


def sha256(value):
    """
    sha256加密
    return:加密结果转成16进制字符串形式
    """
    hash_obj = hashlib.sha256()
    hash_obj.update(value.encode("utf-8"))
    return hash_obj.hexdigest()


def pwd_generator(num=32):
    """
    密码生成器
    :param num: 密码位数
    :return:
    """
    letters = string.ascii_letters + string.digits + "@#%&+="
    key = random.sample(letters, num)
    pwd = "".join(key)
    return pwd


if __name__ == "__main__":
    print(pwd_generator(32))
    # api_key = "7uGt@UwQofA=DO4k6F%+&PibIszV9X2m"
    # api_secret = "bp2q=jtJa8P#dHOzLKFBGg6n1UNmwZTy"
    # req_id = "157078081781"
    # unionText = api_key + req_id + api_secret
    # print("加密前:{}".format(unionText))
    # encryptedText = sha256(unionText)
    # print("加密后:{}".format(encryptedText))



from __future__ import print_function
import json
import os
import io
import os.path as osp
import numpy as np
import PIL.Image
import PIL.ImageDraw
from PIL import Image
import matplotlib.pyplot as plt
import glob
import traceback
from pathlib import Path


def label_colormap(N=256):
    def bitget(byteval, idx):
        return ((byteval & (1 << idx)) != 0)

    cmap = np.zeros((N, 3))
    for i in range(0, N):
        id = i
        r, g, b = 0, 0, 0
        for j in range(0, 8):
            r = np.bitwise_or(r, (bitget(id, 0) << 7 - j))
            g = np.bitwise_or(g, (bitget(id, 1) << 7 - j))
            b = np.bitwise_or(b, (bitget(id, 2) << 7 - j))
            id = (id >> 3)
        cmap[i, 0] = r
        cmap[i, 1] = g
        cmap[i, 2] = b
    cmap = cmap.astype(np.float32) / 255
    return cmap


def polygons_to_mask(img_shape, polygons):
    mask = np.zeros(img_shape[:2], dtype=np.uint8)
    mask = PIL.Image.fromarray(mask)
    xy = list(map(tuple, polygons))
    PIL.ImageDraw.Draw(mask).polygon(xy=xy, outline=1, fill=1)
    mask = np.array(mask, dtype=bool)
    return mask


# similar function as skimage.color.label2rgb
def label2rgb(lbl, img=None, n_labels=None, alpha=0.5, thresh_suppress=0):
    if n_labels is None:
        n_labels = len(np.unique(lbl))

    cmap = label_colormap(n_labels)
    cmap = (cmap * 255).astype(np.uint8)

    lbl_viz = cmap[lbl]
    lbl_viz[lbl == -1] = (0, 0, 0)  # unlabeled

    if img is not None:
        img_gray = PIL.Image.fromarray(img).convert('LA')
        img_gray = np.asarray(img_gray.convert('RGB'))
        # img_gray = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
        # img_gray = cv2.cvtColor(img_gray, cv2.COLOR_GRAY2RGB)
        lbl_viz = alpha * lbl_viz + (1 - alpha) * img_gray
        lbl_viz = lbl_viz.astype(np.uint8)

    return lbl_viz


def draw_label(label, img=None, label_names=None, colormap=None):
    backend_org = plt.rcParams['backend']
    plt.switch_backend('agg')
    plt.rcParams['font.family'] = ['SimHei']
    plt.subplots_adjust(left=0, right=1, top=1, bottom=0,
                        wspace=0, hspace=0)
    plt.margins(0, 0)
    plt.gca().xaxis.set_major_locator(plt.NullLocator())
    plt.gca().yaxis.set_major_locator(plt.NullLocator())

    if label_names is None:
        label_names = [str(l) for l in range(label.max() + 1)]

    if colormap is None:
        colormap = label_colormap(len(label_names))

    label_viz = label2rgb(label, img, n_labels=len(label_names))
    plt.imshow(label_viz)
    plt.axis('off')

    plt_handlers = []
    plt_titles = []
    for label_value, label_name in enumerate(label_names):
        if label_value not in label:
            continue
        if label_name.startswith('_'):
            continue
        fc = colormap[label_value]
        p = plt.Rectangle((0, 0), 1, 1, fc=fc)
        plt_handlers.append(p)
        plt_titles.append('{value}: {name}'
                          .format(value=label_value, name=label_name))
    plt.legend(plt_handlers, plt_titles, loc='lower right', framealpha=.5)

    f = io.BytesIO()
    plt.savefig(f, bbox_inches='tight', pad_inches=0)
    plt.cla()
    plt.close()

    plt.switch_backend(backend_org)

    out_size = (label_viz.shape[1], label_viz.shape[0])
    out = PIL.Image.open(f).resize(out_size, PIL.Image.BILINEAR).convert('RGB')
    out = np.asarray(out)
    return out


def shapes_to_label_sorted(img_shape, shapes, label_name_to_value, type='class'):
    assert type in ['class', 'instance']
    instance_names = []
    ins = None
    cls_name = None
    ins_id = 0
    mask = None
    cls = np.zeros(img_shape[:2], dtype=np.int32)
    if type == 'instance':
        ins = np.zeros(img_shape[:2], dtype=np.int32)
        instance_names = ['_background_']
    tmp = []
    for i in range(0, len(shapes)):
        shape = shapes[i]
        polygons = shape['points']
        label = shape['label']
        if type == 'class':
            cls_name = label
        elif type == 'instance':
            cls_name = label.split('-')[0]
            if label not in instance_names:
                instance_names.append(label)
            ins_id = len(instance_names) - 1
        if cls_name not in label_name_to_value:
            continue
        cls_id = label_name_to_value[cls_name]
        mask = polygons_to_mask(img_shape[:2], polygons)
        num = np.sum(mask.reshape(-1) == True)
        tmp.append((num, mask, cls_id))
        if type == 'instance':
            ins[mask] = ins_id
    # 降序排序
    sort_tmp = sorted(tmp, key=lambda v: v[0], reverse=True)
    for i in range(0, len(sort_tmp)):
        info = sort_tmp[i]
        cls[info[1]] = info[2]
        if type == 'instance':
            ins[mask] = ins_id
    if type == 'instance':
        return cls, ins
    return cls


def labelme_convert_seg(labelme_dir: Path, dist_dir: Path, fetch_labels: list = None):
    """
    labelme转segmentation
    :param labelme_dir:
    :param dist_dir:
    :param fetch_labels: ["汽车", "_background_"]
    :return:
    """
    if not dist_dir.exists():
        dist_dir.mkdir()
    save_img_dir = dist_dir.joinpath("images")
    save_img_vis_dir = dist_dir.joinpath("images_vis")
    save_labels_dir = dist_dir.joinpath("labels")
    if not save_img_dir.exists():
        save_img_dir.mkdir()
    if not save_labels_dir.exists():
        save_labels_dir.mkdir()
    if not save_img_vis_dir.exists():
        save_img_vis_dir.mkdir()

    # get the all class names for the given dataset
    class_names = ['_background_']
    for label_file in glob.glob(osp.join(labelme_dir, '*.json')):
        with open(label_file, encoding="utf-8") as f:
            data = json.load(f)
            for shape in data['shapes']:
                label = shape['label']
                cls_name = label
                if cls_name not in class_names:
                    class_names.append(cls_name)
    class_name_to_id = {}
    class_names_list = []
    class_index = 0
    for i, class_name in enumerate(class_names):
        if fetch_labels:
            if class_name not in fetch_labels:
                continue
        class_id = class_index  # starts with 0
        class_name_to_id[class_name] = class_id
        if class_id == 0:
            assert class_name == '_background_'
        class_index += 1
        class_names_list.append(class_name)
    class_names = tuple(class_names_list)
    print('class_names:', class_names)

    out_class_names_file = osp.join(str(dist_dir), 'labels.txt')
    with open(out_class_names_file, 'w') as f:
        f.writelines('\n'.join(class_names))
    print('Saved class_names:', out_class_names_file)

    colormap = label_colormap(255)
    for root, dirs, files in os.walk(labelme_dir):
        for label_file in files:
            if not label_file.endswith('.json'):
                continue
            print('Generating dataset from:', label_file)
            label_file = os.path.join(root, label_file)
            base = osp.splitext(osp.basename(label_file))[0]
            out_img_file = save_img_dir.joinpath(base + '.jpg')
            out_lbl_file = save_labels_dir.joinpath(base + '.png')
            out_viz_file = save_img_vis_dir.joinpath(base + '.png')
            if out_lbl_file.exists():
                continue
            try:
                with open(label_file, encoding="utf-8") as f:
                    data = json.load(f)
                    img_file = osp.join(osp.dirname(label_file), data['imagePath'])
                    img = PIL.Image.open(img_file)
                    if img.mode != "RGB":
                        img = img.convert('RGB')
                    img = np.asarray(img)
                    PIL.Image.fromarray(img).save(str(out_img_file))
                    # 对标注的label面积由大到小排序，防止因标注顺序问题导致大的遮盖了小的
                    lbl = shapes_to_label_sorted(
                        img_shape=img.shape,
                        shapes=data['shapes'],
                        label_name_to_value=class_name_to_id,
                    )
                    lbl_pil = PIL.Image.fromarray(lbl)
                    # Only works with uint8 label
                    # lbl_pil = PIL.Image.fromarray(lbl, mode='P')
                    # lbl_pil.putpalette((colormap * 255).flatten())
                    lbl_pil.save(str(out_lbl_file))
                    # 生成验证图片-训练不需要，可以屏蔽
                    # label_names = ['%d: %s' % (class_name_to_id.get(cls_name), cls_name) for cls_name in class_name_to_id]
                    label_names = ['%s' % cls_name for cls_name in class_name_to_id]
                    viz = draw_label(lbl, img, label_names, colormap=colormap)
                    PIL.Image.fromarray(viz).save(out_viz_file)
            except Exception as ex:
                print(traceback.print_exc())
                print('程序中断:类别 %s 不在类别列表中' % str(ex))


def check_gray_image(gray_label_file: Path):
    """
    检测是否是灰度标注图片，在图像中存在3个类别时，输出应该是[0, 1, 2]，若上述条件不满足，则说明标注图存在问题，不能直接用于模型训练
    :param gray_label_file:
    :return:
    """
    print(np.unique(np.asarray(Image.open(str(gray_label_file)))))


if __name__ == '__main__':
    pass



# -*- coding:utf-8 -*-
# Author:  zhousf
# Description:
import re

from PIL import ImageFont
from pathlib import Path

# 宋体-汉字
Font_SimSun: Path = Path(__file__).parent.joinpath("SimSun.ttf")

# 符号、数字、字母、数学符号： ①ⓒԑ
Font_SYMBOLA: Path = Path(__file__).parent.joinpath("Symbola.ttf")


def dynamic_font(text: str, font_size: int = 10):
    pattern = re.compile(r'[ⒶⒷⒸⒹⒺⒻⒼⒽⒾⒿⓀⓁⓂⓃⓄⓆⓇⓈⓉⓊⓋⓌⓍⓎⓏⓐⓑⓒⓓⓔⓕⓖⓗⓘⓙⓚⓛⓜⓝⓞⓟⓠⓡⓢⓣⓤⓥⓦⓧⓨⓩ⍺ԑ⍴⍵⍬⍶⍷⍸⍹]')
    if pattern.search(text):
        font = Font_SYMBOLA
    else:
        font = Font_SimSun
    return ImageFont.truetype(font=str(font), size=font_size)



# -*- coding:utf-8 -*-
# Author:      zhousf
# File:        except_util.py
# Description: 非中断流程异常处理：防止函数因异常中断导致后续流程不执行
import warnings


def except_handler(function):
    def catch_except(*args, **key_args):
        try:
            return function(*args, **key_args)
        except Exception as e:
            warnings.warn('{0}: {1}'.format(function.__name__, e))

    return catch_except





# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function: 线程安全的单例模式
import threading


class Singleton(type):
    """
    class A(metaclass=Singleton)
    """
    _instance_lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            with Singleton._instance_lock:
                if not hasattr(cls, "_instance"):
                    cls._instance = super(Singleton,cls).__call__(*args, **kwargs)
        return cls._instance



# -*- coding:utf-8 -*-
# Author:      zhousf
# Description:


def is_empty(dic):
    if dic is None or dic == '':
        return True
    if isinstance(dic, str):
        dic = eval(dic)
    if isinstance(dic, dict):
        if len(dic) != 0:
            return False
    if isinstance(dic, list):
        if len(dic) != 0:
            return False
    if isinstance(dic, set):
        if len(dic) != 0:
            return False
    return True


def pop(dic, key):
    if dic is not None and key is not None:
        if key in dic:
            dic.pop(key)
            return True
    return False


def merge(a, b):
    """
    字典合并，若key重复则b值覆盖a值
    :param a:
    :param b:
    :return:
    example:
    a = {'a':1,'b':{'v':1,'e':2},'g':9}
    b = {'b':{'v':2,'d':2},'c':3,'g':10}
    output:
    {'a': 1, 'c': 3, 'b': {'e': 2, 'd': 2, 'v': 2}, 'g': 10}
    """
    re = {}
    a_dict = a if isinstance(a, dict) else eval(a)
    b_dict = b if isinstance(b, dict) else eval(b)
    key_unit = list(a_dict.keys())
    key_unit.extend(list(b_dict.keys()))
    for key in key_unit:
        if key in a_dict and key in b_dict:
            # 若是字典类型则合并
            if isinstance(a_dict[key], dict):
                # 合并字典
                c = dict(a_dict[key], **b_dict[key])
                re[key] = c
            elif isinstance(a_dict[key], list):
                a_key = {list(k.keys())[0]: list(k.values())[0] for k in a_dict[key]}
                b_key = {list(k.keys())[0]: list(k.values())[0] for k in b_dict[key]}
                for _k in list(b_key.keys()):
                    a_key[_k] = b_key.get(_k)
                r = []
                for _k in list(a_key.keys()):
                    r.append({_k: a_key.get(_k)})
                re[key] = r
            else:
                # 非字典类型则覆盖
                re[key] = b_dict[key]
        elif key in a_dict:
            re[key] = a_dict[key]
        elif key in b_dict:
            re[key] = b_dict[key]
    return re


def merge_list(ml):
    """
    list字典合并
    :param ml:
    :return:
    example：
    a = {'HouYeZiBan-Z': {'aoxian': 0.0118, 'guaca': 0.1205}}
    b = {'HouMen-Z': {'guaca': 0.0505}, 'HouYeZiBan-Z': {'guaca': 0.0003, 'aoxian': 0.0047}}
    c = [a,b]
    merge_list(c)
    output：
    {'HouYeZiBan-Z': {'aoxian': 0.0047, 'guaca': 0.0003}, 'HouMen-Z': {'guaca': 0.0505}}
    """
    if len(ml) < 1:
        return None
    first = ml[0]
    if len(ml) > 1:
        for i in range(1, len(ml)):
            first = merge(first,ml[i])
    return first




# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function:
import openpyxl
from pathlib import Path
from zhousflib.util import re_util


def unmerge_and_fill_cells(excel_file: Path, delete_duplicates_rate: float = 1.0, tmp_excel: Path = None,
                           target_sheet_name=None) -> Path:
    """
    拆分合并单元格并填充有效值
    :param excel_file:
    :param delete_duplicates_rate: 对拆分合并单元格的结果去重的比例，默认为1.0（全相同时去重）,0则不去重
    :param tmp_excel: 临时文件，若为空则会更新源文件
    :param target_sheet_name: None第一张表
    :return:
    """
    wb = openpyxl.load_workbook(str(excel_file))
    contain_merge_cells = False
    for sheet_name in wb.sheetnames:
        if target_sheet_name:
            if target_sheet_name != sheet_name:
                continue
        worksheet = wb[sheet_name]
        all_merged_cell_ranges = list(worksheet.merged_cells.ranges)
        rows_deal = {}
        """
        拆分合并单元格
        """
        for merged_cell_range in all_merged_cell_ranges:
            merged_cell = merged_cell_range.start_cell
            worksheet.unmerge_cells(range_string=merged_cell_range.coord)
            start, end = merged_cell_range.coord.split(":")
            start = int(re_util.get_digit_char(start))
            end = int(re_util.get_digit_char(end))
            if (start, end) not in rows_deal:
                rows_deal[(start, end)] = 1
            else:
                rows_deal[(start, end)] += 1

            for row_index, col_index in merged_cell_range.cells:
                cell = worksheet.cell(row=row_index, column=col_index)
                cell.value = merged_cell.value
        """
        找到符合拆分合并单元格条件的单元格rows
        """
        need_fill = []
        for i in rows_deal:
            need_fill.append(i)
            contain_merge_cells = True
        if len(need_fill) > 0:
            need_fill.sort(key=lambda x: x[0], reverse=False)
        """
        拆分合并单元格后，对空单元格赋予有效值，仅对两个的合并单元格
        """
        for cells in worksheet.iter_rows():
            for cell in cells:
                row = cell.row
                column = cell.column
                for fill in need_fill:
                    count = rows_deal.get(fill)
                    if not count:
                        continue
                    if row == fill[0] and abs(fill[1] - fill[0]) == 1:
                        next_cell = worksheet.cell(row=fill[1], column=column)
                        if not cell.value and next_cell.value:
                            cell.value = next_cell.value
                        rows_deal.pop(fill)
        """
        拆分合并单元格后会有重复的两条，这里去重一下
        """
        if delete_duplicates_rate > 0:
            # 偏移量，记录删除row的个数
            offset = 0
            for fill in need_fill:
                for i in range(fill[0], fill[1]+1):
                    current_data = []
                    next_data = []
                    for row_cells in worksheet.iter_rows(min_row=i-offset, max_row=i-offset):
                        current_data = [cell.value for cell in row_cells]
                    if i < fill[1]:
                        for row_cells in worksheet.iter_rows(min_row=i+1-offset, max_row=i+1-offset):
                            next_data = [cell.value for cell in row_cells]
                    if len(next_data) > 0 and len(current_data) > 0:
                        # 若下一行完全包含上一行或占比满足条件，则删除上一行
                        can_overwrite = True
                        same_cell_count = 0
                        available_cell_count = 0
                        for cell in current_data:
                            if not cell:
                                continue
                            available_cell_count += 1
                            if cell not in next_data:
                                can_overwrite = False
                            else:
                                same_cell_count += 1
                        if can_overwrite or same_cell_count >= delete_duplicates_rate * available_cell_count:
                            worksheet.delete_rows(idx=i-offset)
                            offset += 1
    if tmp_excel:
        wb.save(str(tmp_excel))
        wb.close()
        return tmp_excel
    else:
        if contain_merge_cells:
            wb.save(str(excel_file))
        wb.close()
        return excel_file



# -*- coding:utf-8 -*-
# Author:  zhousf
# Description:  list交集、并集、差集运算
import random


def random_choices(data_list: list, choose_k=3) -> list:
    """
    从列表中随机抽取choose_k个数（会有重复值）
    :param data_list:
    :param choose_k:
    :return:
    """
    return random.choices(data_list, k=choose_k)


def none_filter(data: list) -> list:
    """
    去掉list中的None值
    :param data:
    :return:
    """
    if isinstance(data, list):
        res = []
        for item in data:
            if isinstance(item, list):
                res.append(list(filter(None, item)))
            else:
                res = list(filter(None, data))
                break
        return res
    return data


def intersection(a, b):
    """
    交集
    :param a: [1, 2, 3, 4, 5]
    :param b: [2, 3, 9]
    :return: [2, 3]
    """
    if len(a) == 0:
        return b
    if len(b) == 0:
        return a
    return list(set(a).intersection(set(b)))


def union(a, b):
    """
    并集
    :param a: [1, 2, 3, 4, 5]
    :param b: [2, 3, 9]
    :return: [1, 2, 3, 4, 5, 9]
    """
    if len(a) == 0:
        return b
    if len(b) == 0:
        return a
    return list(set(a).union(set(b)))


def difference(a, b):
    """
    差集
    :param a: [1, 2, 3, 4, 5]
    :param b: [2, 3, 9]
    :return: [1, 4, 5]
    """
    if len(a) == 0:
        return b
    if len(b) == 0:
        return a
    return list(set(a).difference(set(b)))





# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function: Web服务响应实体类
from flask import jsonify


class Response(object):
    log_msg = ""

    def __init__(self, status: int = 200, result="", message: str = ""):
        """
        :param status: 状态
        :param result: 业务处理结果
        :param message: 信息
        """
        self.status = status
        self.message = message if message is not None else message
        self.result = result if result is not None else result

    def log(self, msg):
        """
        记录日志
        :param msg: 日志信息
        :return:
        """
        self.log_msg += "{0}\n".format(msg)


def failed_tip(status: int = 408, result="", msg: str = "failed"):
    """
    业务处理失败
    :param status: 状态
    :param result: 业务处理结果
    :param msg: 信息
    :return:
    """
    info = Response(status=status, result=result, message=msg)
    return jsonify(info.__dict__)


def success_tip(status: int = 200, result="", msg: str = "successful"):
    """
    业务处理成功
    :param status: 状态
    :param result: 业务处理结果
    :param msg: 信息
    :return:
    """
    info = Response(status=status, result=result, message=msg)
    return jsonify(info.__dict__)



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function:
import re


def get_digit_char(string: str):
    # 提取数字
    return re.sub(u"([^\u0030-\u0039])", "", string)


def only_contain_letter_char(self, string: str):
    """
    仅包含字母
    """
    return len(self.get_letter_char(string)) == len(string)


def get_letter_char(string: str):
    # 提取大小写字母
    return re.sub(u"([^\u0041-\u005a\u0061-\u007a])", "", string)


def get_digit_letter_char(string: str):
    # 提取大小写字母、数字
    return re.sub(u"([^\u0041-\u005a\u0061-\u007a\u0030-\u0039])", "", string)